# Configuration file for Sphinx,
# see https://www.sphinx-doc.org/en/master/usage/configuration.html

import os

# -- Recommended settings -----------------------------------------------------

html_theme = "insipid"

html_permalinks_icon = "#"

# If False, source links to Github/GitLab are shown (using html_context)
html_copy_source = False

# Update this with your own repo information:
html_context = {
    "display_github": True,
    "github_user": "kaushikcfd",
    "github_repo": "pybliss",
    "conf_py_path": "/docs/",
    "READTHEDOCS": os.environ.get("READTHEDOCS", False),
    # The 'commit' field is overwritten below (with data from Git).
    # If you prefer, you can also get it from your build environment,
    # e.g. READTHEDOCS_GIT_COMMIT_HASH.
    "commit": "???",
}

# -- Settings for source code -------------------------------------------------

# Language used for syntax highlighting (default: 'python')
# highlight_language = 'none'

# Style of syntax highlighting
# pygments_style = 'monokai'

# -- Language settings --------------------------------------------------------

# language = 'es'

# -- Theme configuration ------------------------------------------------------

html_theme_options = {
    # 'body_centered': False,
    # 'body_max_width': None,
    # 'breadcrumbs': True,
    # 'enable_search_shortcuts': False,
    # 'globaltoc_collapse': False,
    # 'globaltoc_includehidden': True,
    # 'initial_sidebar_visibility_threshold': None,
    # 'left_buttons': [
    # ],
    # 'navigation_with_keys': False,
    # 'nosidebar': True,
    # 'right_buttons': [
    #    'search-button.html',
    # ],
    # 'rightsidebar': True,
    # 'show_insipid': False,
    # 'sidebar_overlay_width': None,
    # 'sidebar_transition': '1s ease-out',
    # 'sidebarwidth': '10rem',
    # 'strip_section_numbers': False,
    # 'topbar_transition': '1.5s ease-out',
}

html_sidebars = {
   "**": [
       "home.html",
       "globaltoc.html",
       "separator.html",
       "indices.html",
   ],
   "showcase/different-sidebar": [
       "localtoc.html",
       "searchbox.html",
   ],
   "showcase/no-sidebar": [],
}

# -- Project information ------------------------------------------------------

project = "insipid-sphinx-theme"
# html_title = 'Insipid Sphinx Theme'
html_short_title = "insipid"
version = "2025.4"


# -- Page footer --------------------------------------------------------------

html_show_copyright = False
# html_show_sphinx = False
# html_show_sourcelink = False

# Only relevant when html_copy_source is True
# html_sourcelink_suffix = ''

# -- Miscellaneous settings ---------------------------------------------------

# Numbered figures, tables and code-blocks
numfig = True

html_secnumber_suffix = "\N{FIGURE SPACE}"

# html_compact_lists = False

# smartquotes = False

# Generate alphabetic index
# html_use_index = False

# Separate page per starting letter
# html_split_index = True

# Generate domain indices, e.g. Python module index
# html_domain_indices = False

# The default in Sphinx 4+ is 'inline'
# html_codeblock_linenos_style = 'table'

# -- Sphinx extensions --------------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
    "nbsphinx",
]

intersphinx_mapping = {
    "IPython": ("https://ipython.readthedocs.io/en/stable/", None),
    "matplotlib": ("https://matplotlib.org/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "pytools": ("https://documen.tician.de/pytools/", None),
}



import numpy as np

import pybliss as bliss


def test_create_graph():
    g = bliss.Graph(42)
    assert g.nvertices == 42


def test_dimacs_conversion():
    pentagon = bliss.Graph(5)
    for i in range(5):
        pentagon.add_edge(i, (i + 1) % 5)

    with open("pentagon.dimacs", "w") as fp:
        pentagon.write_dimacs(fp)

    with open("pentagon.dimacs") as fp:
        assert bliss.Graph.from_dimacs(fp) == pentagon


def test_to_dot():
    pentagon = bliss.Graph(5)
    for i in range(5):
        pentagon.add_edge(i, (i + 1) % 5)

    dot_str = pentagon.to_dot()
    assert isinstance(dot_str, str)
    assert dot_str


def test_write_dot():
    pentagon = bliss.Graph(5)
    for i in range(5):
        pentagon.add_edge(i, (i + 1) % 5)

    with open("pentagon.dot", "w") as fp:
        pentagon.write_dot(fp)

    with open("pentagon.dot") as fp:
        assert fp.read() == pentagon.to_dot()


def test_pentagram_pentagon_canon():
    pentagon = bliss.Graph(5)
    for i in range(5):
        pentagon.add_edge(i, (i + 1) % 5)

    pentagram = bliss.Graph(5)
    for i in range(5):
        pentagram.add_edge(i, (i + 2) % 5)

    assert pentagon != pentagram
    s1 = bliss.Stats()
    s2 = bliss.Stats()
    pentagon_perm = pentagon.get_permutation_to_canonical_form(s1)
    pentagram_perm = pentagram.get_permutation_to_canonical_form(s2)
    pentagon_canon = pentagon.permute(pentagon_perm)
    pentagram_canon = pentagram.permute(pentagram_perm)
    assert pentagon_canon == pentagram_canon


def test_find_automorphisms_report():
    # Let's explore the automorphisms of the petersen graph
    petersen = bliss.Graph(10)

    for i in range(5):
        # Outer 5-cycle (0-1-2-3-4-0)
        petersen.add_edge(i, (i + 1) % 5)
        # Inner star-shaped connections (5-7-9-6-8-5)
        petersen.add_edge(i + 5, ((i + 2) % 5) + 5)
        # Connections between outer and inner nodes (0-5, 1-6, 2-7, 3-8, 4-9)
        petersen.add_edge(i, i + 5)

    n_generator = 0

    def report(n, aut):
        nonlocal n_generator
        n_generator += 1
        assert isinstance(aut, np.ndarray)
        assert isinstance(n, int)
        print(f"Found a new automorphism generator of length {n}, aut = {aut}.")

    s = bliss.Stats()
    petersen.find_automorphisms(s, report)

    # petersen graph has 3 automorphism generators
    assert n_generator == 3

    # petersen graph has |Aut| = 120
    assert s.group_size == 120


def test_find_automorphisms_report_terminate():
    # Let's explore the automorphisms of the petersen graph
    petersen = bliss.Graph(10)

    for i in range(5):
        # Outer 5-cycle (0-1-2-3-4-0)
        petersen.add_edge(i, (i + 1) % 5)
        # Inner star-shaped connections (5-7-9-6-8-5)
        petersen.add_edge(i + 5, ((i + 2) % 5) + 5)
        # Connections between outer and inner nodes (0-5, 1-6, 2-7, 3-8, 4-9)
        petersen.add_edge(i, i + 5)

    def report(n, aut):
        print(f"Found a new automorphism generator of length {n}, aut = {aut}.")

    isearch = 0

    def terminate():
        nonlocal isearch
        if isearch < 10:
            isearch += 1
            return False
        else:
            return True

    s = bliss.Stats()
    petersen.find_automorphisms(s, report, terminate)

    # terminate early
    assert isearch == 10


def test_change_color_get_color():
    k3 = bliss.Graph(3)
    for i in range(3):
        k3.add_edge(i, (i + 1) % 3)
        # set the color of each i-vertex as i.
        k3.change_color(i, i)

    for i in range(3):
        assert k3.get_color(i) == i



import numpy as np

import pybliss as bliss


def test_graph_from_numpy_0():
    pentagon_0 = bliss.Graph(5)
    for i in range(5):
        pentagon_0.add_edge(i, (i + 1) % 5)

    for i in range(5):
        pentagon_0.change_color(i, i**2)

    pentagon_adj = np.array(
        [
            [0, 1, 0, 0, 1],
            [1, 0, 1, 0, 0],
            [0, 1, 0, 1, 0],
            [0, 0, 1, 0, 1],
            [1, 0, 0, 1, 0],
        ]
    )

    pentagon_1 = bliss.utils.graph_from_numpy(
        5, np.argwhere(pentagon_adj), np.arange(5) ** 2
    )

    assert pentagon_0 == pentagon_1


def test_digraph_from_numpy_0():
    pentagon_0 = bliss.Digraph(5)
    for i in range(5):
        pentagon_0.add_edge(i, (i + 1) % 5)

    for i in range(5):
        pentagon_0.change_color(i, i**2)

    pentagon_adj = np.array(
        [
            [0, 1, 0, 0, 0],
            [0, 0, 1, 0, 0],
            [0, 0, 0, 1, 0],
            [0, 0, 0, 0, 1],
            [1, 0, 0, 0, 0],
        ]
    )

    pentagon_1 = bliss.utils.digraph_from_numpy(
        5, np.argwhere(pentagon_adj), np.arange(5) ** 2
    )

    assert pentagon_0 == pentagon_1


def test_graph_to_numpy_0():
    pentagon_0 = bliss.Graph(5)
    for i in range(5):
        pentagon_0.add_edge(i, (i + 1) % 5)

    for i in range(5):
        pentagon_0.change_color(i, i**2)
    pentagon_1 = bliss.utils.graph_from_numpy(
        5, *bliss.utils.graph_to_numpy(pentagon_0)
    )

    assert pentagon_0 == pentagon_1


def test_digraph_to_numpy_0():
    pentagon_0 = bliss.Digraph(5)
    for i in range(5):
        pentagon_0.add_edge(i, (i + 1) % 5)

    for i in range(5):
        pentagon_0.change_color(i, i**2)

    pentagon_1 = bliss.utils.digraph_from_numpy(
        5, *bliss.utils.graph_to_numpy(pentagon_0)
    )

    assert pentagon_0 == pentagon_1



import pybliss as bliss


def test_create_graph():
    g = bliss.Digraph(42)
    assert g.nvertices == 42


def test_dimacs_conversion():
    pentagon = bliss.Digraph(5)
    for i in range(5):
        pentagon.add_edge(i, (i + 1) % 5)

    with open("pentagon.dimacs", "w") as fp:
        pentagon.write_dimacs(fp)

    with open("pentagon.dimacs") as fp:
        assert bliss.Digraph.from_dimacs(fp) == pentagon


def test_to_dot():
    pentagon = bliss.Digraph(5)
    for i in range(5):
        pentagon.add_edge(i, (i + 1) % 5)

    dot_str = pentagon.to_dot()
    assert isinstance(dot_str, str)
    assert dot_str


def test_write_dot():
    pentagon = bliss.Digraph(5)
    for i in range(5):
        pentagon.add_edge(i, (i + 1) % 5)

    with open("pentagon.dot", "w") as fp:
        pentagon.write_dot(fp)

    with open("pentagon.dot") as fp:
        assert fp.read() == pentagon.to_dot()


def test_change_color_get_color():
    k3 = bliss.Digraph(3)
    for i in range(3):
        k3.add_edge(i, (i + 1) % 3)
        # set the color of each i-vertex as i.
        k3.change_color(i, i)

    for i in range(3):
        assert k3.get_color(i) == i

# TODO: Add tests for Digraph.get_caonnical_from, Digraph.find_automorphisms.



from __future__ import annotations

from typing import TypeVar

import numpy as np

from .pybliss_ext import (
    Digraph,
    Graph,
    permutation_to_str,
    print_permutation_to_file,
)

__doc__ = """
.. autofunction:: permutation_to_str
.. autofunction:: print_permutation_to_file
.. autofunction:: graph_from_numpy
.. autofunction:: digraph_from_numpy
.. autofunction:: graph_to_numpy
.. autofunction:: digraph_to_numpy
"""


__all__ = [
    "digraph_from_numpy",
    "graph_from_numpy",
    "permutation_to_str",
    "print_permutation_to_file",
]

GT = TypeVar("GT", Graph, Digraph)


def _from_numpy_checks(
    N: int,
    E: np.ndarray[tuple[int, int], np.dtype[np.integer]],
    c: np.ndarray[tuple[int], np.dtype[np.integer]],
) -> None:
    assert N >= 0
    assert E.ndim == 2, ""
    assert c.ndim == 1, "'c' must be a column vector."
    assert (
        c.shape[0] == N
    ), "'c' should have a color for every vertex in the graph."


def _graph_or_digraph_from_numpy(
    G: GT,
    E: np.ndarray[tuple[int, int], np.dtype[np.integer]],
    c: np.ndarray[tuple[int], np.dtype[np.integer]],
) -> GT:
    for i, j in E:
        G.add_edge(i, j)

    for i, color in enumerate(c):
        G.change_color(i, color)

    return G


def graph_from_numpy(
    N: int,
    E: np.ndarray[tuple[int, int], np.dtype[np.integer]],
    c: np.ndarray[tuple[int], np.dtype[np.integer]],
) -> Graph:
    r"""
    Returns a :class:`pybliss.Graph`, *G*, with *N*-vertices, *E*-edges and
    colored vector *c*.

    :arg N: The number of vertices in the graph, G.
    :arg E: A 2D :class:`numpy.ndarray` of shape
        :math:`N_{\mathrm{edges}} \times 2`, where :math:`N_{\mathrm{edges}}` is
        the number of edges in G. Each row in *E* is the of the form ``[i, j]``
        to denote that there is an edge from ``i`` to ``j`` in *G*.
    :arg c: An *N*-long column vector such that ``c[i]`` corresponds to the
        color of i-th vertex in G.
    """
    _from_numpy_checks(N, E, c)
    G = Graph(N)
    _graph_or_digraph_from_numpy(G, E, c)
    return G


def digraph_from_numpy(
    N: int,
    E: np.ndarray[tuple[int, int], np.dtype[np.integer]],
    c: np.ndarray[tuple[int], np.dtype[np.integer]],
) -> Digraph:
    r"""
    Returns a :class:`pybliss.Digraph`, *G*, with *N*-vertices, *E*-edges and
    colored vector *c*.

    :arg N: The number of vertices in the graph, G.
    :arg E: A 2D :class:`numpy.ndarray` of shape
        :math:`N_{\mathrm{edges}} \times 2`, where :math:`N_{\mathrm{edges}}` is
        the number of edges in G. Each row in *E* is the of the form ``[i, j]``
        to denote that there is an edge from ``i`` to ``j`` in *G*.
    :arg c: An *N*-long column vector such that ``c[i]`` corresponds to the
        color of i-th vertex in G.
    """
    _from_numpy_checks(N, E, c)
    G = Digraph(N)
    _graph_or_digraph_from_numpy(G, E, c)
    return G


def _parse_dimacs(
    dimacs_src: str,
) -> tuple[
    np.ndarray[tuple[int, int], np.dtype[np.integer]],
    np.ndarray[tuple[int], np.dtype[np.integer]],
]:
    # See <https://users.aalto.fi/~tjunttil/bliss/fileformat.html>
    lines = dimacs_src.split("\n")
    lines = [line for line in lines if line]
    istart = 0
    while True:
        if not lines:
            break
        if not lines[istart].startswith("c"):
            break
        assert lines[istart].startswith("c")
        istart += 1

    problem_defn_line = lines[istart]
    assert problem_defn_line.startswith("p edge ")
    N, E = [
        int(val)
        for val in problem_defn_line[len("p edge ") :].strip().split(" ")
    ]
    istart += 1

    edges = np.empty((E, 2), dtype=np.uint64)
    colors = np.zeros(N, dtype=np.uint64)

    i_edge = 0

    for i in range(istart, len(lines)):
        if lines[i].startswith("n"):
            node_i, c = [int(val) for val in lines[i][2:].strip().split(" ")]
            assert c >= 0
            colors[node_i - 1] = c
        elif lines[i].startswith("e"):
            node_i, node_j = [
                int(val) for val in lines[i][2:].strip().split(" ")
            ]
            edges[i_edge, 0] = node_i - 1
            edges[i_edge, 1] = node_j - 1
            i_edge += 1
        else:
            raise AssertionError(f"Ill-formed dimacs lines: '{lines[i]}'?")

    assert i_edge == E

    return edges, colors


def graph_to_numpy(
    G: Graph,
) -> tuple[
    np.ndarray[tuple[int, int], np.dtype[np.integer]],
    np.ndarray[tuple[int], np.dtype[np.integer]],
]:
    r"""
    Returns a tuple ``(E, c)`` for an undirected graph G, where:

    - **E** is a 2D :class:`numpy.ndarray` of shape
      :math:`N_{\mathrm{edges}} \times 2`, where :math:`N_{\mathrm{edges}}` is
      the number of edges in *G*. Each row of *E* has the form ``[i, j]``,
      indicating that there is an edge from vertex ``i`` to vertex ``j`` in *G*.
    - **c** is a 1D :class:`numpy.ndarray` of length :math:`N` such that
      ``c[i]`` corresponds to the color of the *i*-th vertex in *G*.
    """
    return _parse_dimacs(G.to_dimacs())


def digraph_to_numpy(
    G: Digraph,
) -> tuple[
    np.ndarray[tuple[int, int], np.dtype[np.integer]],
    np.ndarray[tuple[int], np.dtype[np.integer]],
]:
    r"""
    Returns a tuple ``(E, c)`` for an undirected graph G, where:

    - **E** is a 2D :class:`numpy.ndarray` of shape
      :math:`N_{\mathrm{edges}} \times 2`, where :math:`N_{\mathrm{edges}}` is
      the number of edges in *G*. Each row of *E* has the form ``[i, j]``,
      indicating that there is an edge from vertex ``i`` to vertex ``j`` in *G*.
    - **c** is a 1D :class:`numpy.ndarray` of length :math:`N` such that
      ``c[i]`` corresponds to the color of the *i*-th vertex in *G*.
    """
    return _parse_dimacs(G.to_dimacs())
